/* Copyright (C) 2010 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package backend.arm;

import compbase.Mthd;
import compbase.Ops;
import compbase.StdTypes;

import backend.Architecture;
import backend.Instruction;

/**
 * ARM7: Architecture backend for ARM7 processors
 * 
 * @author S. Frenz
 * @version 100929 added basic register management
 *  version 100805 initial version
 */

public class ARM7 extends Architecture {
  private final static int RegCLSS = 0;
  private final static int RegINST = 1;
  private final static int RegBASE = 12;
  private final static int RegALL = 0x00000000; //mask for general purpose registers
  
  private int usedRegs, writtenRegs;
  
  public ARM7() {
    relocBytes=4;
    stackClearBits=3;
    allocClearBits=3;
    maxInstrCodeSize=8;
    throwFrameSize=0;
    throwFrameExcOff=0;
    regClss=RegCLSS;
    regInst=RegINST;
    regBase=RegBASE;
    binAriCall[StdTypes.T_BYTE]|=(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_SHRT]|=(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_INT]|=(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_LONG]|=(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_FLT]|=(1<<(Ops.A_MUL-Ops.MSKBSE))|(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE))|(1<<(Ops.A_PLUS-Ops.MSKBSE))|(1<<(Ops.A_MINUS-Ops.MSKBSE));
    binAriCall[StdTypes.T_DBL]|=(1<<(Ops.A_MUL-Ops.MSKBSE))|(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE))|(1<<(Ops.A_PLUS-Ops.MSKBSE))|(1<<(Ops.A_MINUS-Ops.MSKBSE));
    unaAriCall[StdTypes.T_FLT]|=(1<<(Ops.A_MINUS-Ops.MSKBSE));
    unaAriCall[StdTypes.T_DBL]|=(1<<(Ops.A_MINUS-Ops.MSKBSE));
  }
  
  public Instruction createNewInstruction() {
    return new ARMInstr(maxInstrCodeSize);
  }
  
  public void putRef(Object loc, int offset, Object ptr, int ptrOff) {
    mem.putInt(loc, offset, mem.getAddrAsInt(ptr, ptrOff));
  }
  
  public void putCodeRef(Object loc, int offset, Object ptr, int ptrOff) {
    // TODO Auto-generated method stub
    mem.putInt(loc, offset, mem.getAddrAsInt(ptr, ptrOff));
  }
  
  protected int bitSearch(int value, int hit, int prefere) {
    int i, j;
    
    //first: try to search in prefered ones
    prefere&=value;
    if (prefere!=0 && prefere!=value) for (i=0; i<32; i++) { //search here only if there is a difference
      j=1<<i;
      if ((prefere&j)!=0) {
        if (--hit==0) return j;
      }
    }
    //nothing prefered, try in all possible
    for (i=0; i<32; i++) {
      j=1<<i;
      if ((value&j)!=0) {
        if (--hit==0) return j;
      }
    }
    //nothing found, sorry
    return 0;
  }
  
  protected int freeRegSearch(int mask, int type) {
    int ret, ret2;
    
    mask&=RegALL;
    if ((ret=bitSearch(mask, 1, 0))==0) return 0;
    if (type==StdTypes.T_LONG || type==StdTypes.T_DPTR) {
      if ((ret2=bitSearch(mask&~ret, 1, 0))==0) return 0;
      ret|=ret2;
    }
    return ret;
  }
  
  protected int storeReg(int regs) {
    regs&=usedRegs&writtenRegs;
    //TODO push remaining regs
    return regs;
  }
  
  protected void restoreReg(int regs) {
    usedRegs|=regs;
    writtenRegs|=regs;
    //TODO pop given regs
  }
  
  public int ensureFreeRegs(int ignoreReg1, int ignoreReg2, int keepReg1, int keepReg2) {
    // TODO Auto-generated method stub
    return 0;
  }
  
  public int prepareFreeReg(int avoidReg1, int avoidReg2, int reUseReg, int type) {
    // TODO Auto-generated method stub
    return 0;
  }
  
  public int allocReg() {
    // TODO Auto-generated method stub
    return 0;
  }
  
  public void deallocRestoreReg(int deallocRegs, int keepRegs, int restoreRegs) {
    // TODO Auto-generated method stub
  }
  
  public void finalizeInstructions(Instruction first) {
    // TODO Auto-generated method stub
  }
  
  public Mthd prepareMethodCoding(Mthd mthd) {
    // TODO Auto-generated method stub
    return null;
  }
  
  public void codeProlog() {
    // TODO Auto-generated method stub
  }
  
  public void codeEpilog(Mthd outline) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadConstVal(int dst, int val, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadConstDoubleOrLongVal(int dst, long val, boolean asDouble) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadVarAddr(int dstReg, int srcReg, Object loc, int off) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadVarVal(int dstReg, int srcReg, Object loc, int off, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genConvertVal(int dst, int src, int toType, int fromType) {
    // TODO Auto-generated method stub
  }
  
  public void genDup(int dst, int src, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genPushConstVal(int val, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genPushConstDoubleOrLongVal(long val, boolean asDouble) {
    // TODO Auto-generated method stub
  }
  
  public void genPush(int src, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genPop(int dst, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genAssign(int dst, int src, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genBinOp(int dst, int src1, int src2, int op, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genUnaOp(int dst, int src, int op, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genIncMem(int dst, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genDecMem(int dst, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genSaveUnitContext() {
    // TODO Auto-generated method stub
  }
  
  public void genRestUnitContext() {
    // TODO Auto-generated method stub
  }
  
  public void genLoadUnitContext(int dst, int off) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadConstUnitContext(int dst, Object unitLoc) {
    // TODO Auto-generated method stub
  }
  
  public void genSaveInstContext() {
    // TODO Auto-generated method stub
  }
  
  public void genRestInstContext() {
    // TODO Auto-generated method stub
  }
  
  public void genLoadInstContext(int src) {
    // TODO Auto-generated method stub
  }
  
  public void genCall(int off, int clssReg, boolean normal, int parSize) {
    // TODO Auto-generated method stub
  }
  
  public void genCallIndexed(int intfReg, int off, int parSize) {
    // TODO Auto-generated method stub
  }
  
  public void genCallConst(Mthd obj, int parSize) {
    // TODO Auto-generated method stub
  }
  
  public void genJmp(Instruction dest) {
    // TODO Auto-generated method stub
  }
  
  public void genCondJmp(Instruction dest, int condHnd) {
    // TODO Auto-generated method stub
  }
  
  public int genComp(int src1, int src2, int type, int cond) {
    // TODO Auto-generated method stub
    return cond;
  }
  
  public int genCompValToConstVal(int src, int val, int type, int cond) {
    // TODO Auto-generated method stub
    return cond;
  }
  
  public int genCompValToConstDoubleOrLongVal(int src, long val, boolean asDouble, int cond) {
    // TODO Auto-generated method stub
    return cond;
  }
  
  public int genCompPtrToNull(int reg, int cond) {
    // TODO Auto-generated method stub
    return cond;
  }
  
  public void genWriteIO(int dst, int src, int type, int memLoc) {
    // TODO Auto-generated method stub
  }
  
  public void genReadIO(int dst, int src, int type, int memLoc) {
    // TODO Auto-generated method stub
  }
  
  public void genMoveToPrimary(int src, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genMoveFromPrimary(int dst, int type) {
    // TODO Auto-generated method stub
  }
  
  public void genMoveIntfMapFromPrimary(int dst) {
    // TODO Auto-generated method stub
  }
  
  public void genSavePrimary(int type) {
    // TODO Auto-generated method stub
  }
  
  public void genRestPrimary(int type) {
    // TODO Auto-generated method stub
  }
  
  public void genCheckBounds(int addrReg, int offReg, int checkToOffset, Instruction onSuccess) {
    // TODO Auto-generated method stub
  }
  
  public void genCheckStackExtreme(int maxValueReg, Instruction onSuccess) {
    // TODO Auto-generated method stub
  }
  
  public void genLoadDerefAddr(int addrReg, int indReg, int baseOffset, int entrySize) {
    // TODO Auto-generated method stub
  }
  
  public void genThrowFrameBuild(int globalAddrReg, Instruction dest, int throwBlockOffset) {
    // TODO Auto-generated method stub
  }
  
  public void genThrowFrameUpdate(Instruction oldDest, Instruction newDest, int throwBlockOffset) {
    // TODO Auto-generated method stub
  }
  
  public void genThrowFrameReset(int globalAddrReg, int throwBlockOffset) {
    // TODO Auto-generated method stub
  }
  
  public void genStoreVarVal(int objReg, Object loc, int off, int valReg, int type) {
    // TODO Auto-generated method stub
    int addr, restore;
    restore=prepareFreeReg(valReg, objReg, 0, StdTypes.T_PTR);
    addr=allocReg();
    genLoadVarAddr(addr, objReg, loc, off);
    genAssign(addr, valReg, type);
    deallocRestoreReg(addr, 0, restore);
  }
  
  public void genStoreVarConstVal(int objReg, Object loc, int off, int val, int type) {
    // TODO Auto-generated method stub
    int valReg, restore;
    restore=prepareFreeReg(objReg, 0, 0, type);
    valReg=allocReg();
    genLoadConstVal(valReg, val, type);
    genStoreVarVal(objReg, loc, off, valReg, type);
    deallocRestoreReg(valReg, 0, restore);
  }
  
  public void genStoreVarConstDoubleOrLongVal(int objReg, Object loc, int off, long val, boolean asDouble) {
    // TODO Auto-generated method stub
    int valReg, restore, type=asDouble ? StdTypes.T_DBL : StdTypes.T_LONG;
    restore=prepareFreeReg(objReg, 0, 0, type);
    valReg=allocReg();
    genLoadConstDoubleOrLongVal(valReg, val, asDouble);
    genStoreVarVal(objReg, loc, off, valReg, type);
    deallocRestoreReg(valReg, 0, restore);
  }
  
  public void genBinOpConstRi(int dst, int src1, int val, int op, int type) { //may destroy src1
    // TODO Auto-generated method stub
    int src2, restore, valType=((op>>>16)==Ops.S_BSH) ? StdTypes.T_INT : type;
    restore=prepareFreeReg(dst, src1, 0, valType);
    src2=allocReg();
    genLoadConstVal(src2, val, valType);
    genBinOp(dst, src1, src2, op, type);
    deallocRestoreReg(src2, 0, restore);
  }
  
  public void genBinOpConstDoubleOrLongRi(int dst, int src1, long val, int op, boolean asDouble) {
    // TODO Auto-generated method stub
    int src2, restore, type=asDouble ? StdTypes.T_DBL : StdTypes.T_LONG;
    restore=prepareFreeReg(dst, src1, 0, type);
    src2=allocReg();
    genLoadConstDoubleOrLongVal(src2, val, asDouble);
    genBinOp(dst, src1, src2, op, type);
    deallocRestoreReg(src2, 0, restore);
  }
  
  public void inlineVarOffset(int inlineMode, int objReg, Object loc, int offset, int baseValue) {
    // TODO Auto-generated method stub
  }
}
